static char IDAgent[] = "@(#)Agent %D";
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#
#NAME
#  Agent = JC's network file daemon
#
#SYNOPSIS
#  cd dir; Agent [name] &
#
#DESCRIPTION
#
#OPTIONS
#
#ENVIRONMENT
#
#BUGS
#
#SEE ALSO
#
#AUTHOR
#  John Chambers <jc@trillian.mit.edu>
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "D.h"
#include "err.h"
#include "str.h"
#include "mem.h"
#include "sys_socket.h"
#include "sys_in.h"
#include "sys_netdb.h"
#include "sys_signal.h"
#include "sys_file.h"
#include "sys_fcntl.h"
#include "sys_stat.h"

#define MAXNAME 256

#define BUFLEN 1024
#define CMDLEN 10		/* Small for debugging */
#define RSPLEN 1024
int    bufmax = BUFLEN;
int    cmdmax = CMDLEN;	/* Small for debugging */
int    rspmax = RSPLEN;

Flag   use_accept = 1;
int    child = -1;		/* Child process's pid */
CP     cmd = 0;			/* Current command name */
int    cons = 0;		/* Number of connections received so far */
int    conlim = 0;		/* Number of connections to accept before exiting */
Flag   connected = 0;	/* Connection verified */
Flag   echo = 0;		/* Echo messages back to client */
Flag   getsocket = 1;	/* Should we open a socket? */
char  *hostname = 0;	/* The hostname to bind to */
char  *myname = 0;		/* Symbolic name to respond to */
char  *ifmt = "%d";		/* How to decode numbers */
int    msgs = 0;		/* Number of messages received so far */
int    msglim = 0;		/* Number of messages to read before exiting */
int    asock = -1;		/* Accept socket */
int    lsock=0;			/* Listening socket */
/*t    mypid = -1;		** This process's pid [now in libV.a] */
TIMV   tmout = {1};		/* How long to wait in select() */
char  *portname = 0;	/* Symbolic version of port number */
int    port = 0;		/* TCP port number */
int    qlen = 5;		/* Length of listen() queue */
U32    now, then;		/* Timestamps */
fd_set rmask = {0};		/* Select read mask */
int    timelimit = 0;	/* Seconds to wait for messages */
SKin   client = {0};	/* Address of client */
int    cliens = 0;		/* Size of client's address */
int    wsleep = 1;		/* Seconds between writes */
Flag   exec = 0;		/* Attempt to exec messages */

global Str** msgp = 0;	/* Current message */
global int   msgf = 0;	/* Current number of fields */
global int   msgm = 2;	/* Maximum number of fields */
global Str   Buf = {0};	/* Input    buffer */
global Str   Cmd = {0};	/* Command  buffer */
global Str   Dat = {0};	/* Data     buffer */
global Str   Rsp = {0};	/* Response buffer */

#ifdef SIGIO
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
Flag  sigiofl = 0;
int   sigiosleep = 100;	/* How long to sleep in SIGIO loop */
Flag  use_sigio = 0;

sigio()
{
/*:1	P2("%s\tSIGIO.",Vtime());*/
	sigiofl = 1;
/*:1	P2("%s\tsigio return ...",Vtime());*/
:9	fprintf(stderr,"sigio() return ...\n"); fflush(stderr);
}
#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
main(ac,av)
	int   ac;
	char**av;
{	int   r=0;		/* Result (exit value) */
	char *p;
	int   c0=0, c1=0;
	int   a=0, i=0, n=0;
	int   length;
	SKin  addr;
	HOST *hp=0, *gethostbyname();

:1	ac = Dinit(ac,av);
	mypid = getpid();
	if (!(msgp = (Str**)GetAChunk(msgm*sizeof(Str*),sizeof(Str*),"Message List"))) {
:1		V1 "*** Can't get %d bytes for Message List [Err %d=%s=%s]",
:1			msgm*sizeof(Str*),Errinfo V;
		Exit(errno);
	}

	for (a=1; a<ac; a++) {
:3		V3 "Arg %2d: \"%s\"",a,av[a] V;
		switch(c0 = av[a][0]) {
		case '-':
		case '+':
:2			V3 "Opt %2d: \"%s\"",a,av[a] V;
			switch (c1 = av[a][1]) {
			case 'a':
			case 'A':
				if (!strcasecmp(av[a]+1,"accept")) {
					use_accept = (c0 == '+');
:2					V2 "%ssing accept.",use_accept?"U":"Not u" V;
					break;
				}
				goto badopt;
			case 'c':
			case 'C':
				if (sscanf(av[a]+2,ifmt,&conlim) < 1)
					++n;
:2				V2 "Accept only %d connections.",conlim V;
				break;
			case 'e':
			case 'E':
				echo = (c0 == '+');
:2				V2 "Echo %sabled.",echo?"en":"dis" V;
				break;
			case 'n':
			case 'N':
				if (sscanf(av[a]+2,ifmt,&msglim) < 1)
					++n;
:2				V2 "Accept only %d messages.",msglim V;
				break;
			case 's':
			case 'S':
#ifdef SIGIO
				if (!strcasecmp(av[a]+1,"SIGIO")) {
					use_sigio = (c0 == '+');
:2					V2 "%ssing SIGIO.",use_sigio?"U":"Not u" V;
					break;
				}
#endif
				if (sscanf(av[a]+2,ifmt,&wsleep) < 1)
					++n;
:2				V2 "Wait %d seconds between messages.",wsleep V;
				break;
			case 't':
			case 'T':
				if (sscanf(av[a]+2,ifmt,&timelimit) < 1)
					++n;
:2				V2 "Wait %d seconds for replies.",timelimit V;
				break;
			case 'x':
			case 'X':
				exec = (c0 == '+');
:2				V2 "Exec %sabled.",exec?"en":"dis" V;
				break;
			badopt:
:1			default:
:1				V1 "Arg %2d: \"%s\" ignored.",a,av[a] V;
:1				break;
			}
			break;
		default:
			if (!portname && (sscanf(av[a],ifmt,&port) > 0)) {
				portname = av[a];
:2				V2 "Port: \"%s\"",portname V;
:3				V3 "port: %d.",port V;
			} else
			if (!myname) {
				myname = av[a];
:2				V2 "myname: \"%s\"",myname V;
:1			} else {
:1				V1 "Arg %2d: \"%s\" ignored.",a,av[a] V;
			}
			break;
		}
	}
	if (!port) port = 4217;
#ifdef SIGIO
	if (use_sigio) {
:3		V3 "Catch SIGIO." V;
		SignalM(SIGIO,sigio,"SIGIO");
	}
#endif
/*
* Create socket from which to read.
*/
	if (getsocket) {
		lsock = Socket(AF_INET, SOCK_STREAM, 0);
		if (lsock < 0) {
:1			V1 "Can't get STREAM socket [Err %d=%s=%s]",Errinfo V;
			r = 2;
			Fail;
		}
/*
* Create address with wildcards.
*/
		if (!hostname) {
			if (!(hostname = (char*)MallocM(MAXNAME,"hostname"))) {
				r = errno;
				Fail;
			}
			if (Gethostname(hostname, MAXNAME)) {
				r = errno;
:1				V1 "Can't get hostname [Err %d=%s=%s]",Errinfo V;
				Fail;
			}
		}
:3		V3 "Host: \"%s\"",hostname V;
:x		hp = Gethostbyn(hostname);
:x		Bcopy(hp->h_addr,&(addr.sin_addr.s_addr),hp->h_length);
:x	} else {
		Bzero(&(addr.sin_addr.s_addr),sizeof(addr.sin_addr.s_addr));
	}
	if (!myname) {
		myname = hostname;
:2		V2 "Set myname=hostname=%s",Dspp(myname) V;
	}
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	if (Vlvl > 3) DmpSKin(&addr);
:6	H6(&addr,sizeof(addr),"addr");
	if (Bind(lsock,&addr,sizeof(addr)) < 0) {
:1		V1 "Can't bind to address %s.%d [Err %d=%s=%s]",
:1			ipdot(&addr.sin_addr),
:1			ntohs( addr.sin_port),
:1			Errinfo V;
		r = 2;
		Fail;
	}
/*
* Find assigned port value and print it out.
*/
	length = sizeof(addr);
	if (Getsockname(lsock, &addr, &length)) {
:1		V1 "Can't get socket address for file %d [Err %d=%s=%s]",lsock,Errinfo V;
		r = 2;
		Fail;
	}
:5	H6(&addr,sizeof(addr),"addr");
	printf("TCP socket has address %s port %d.\n",
		ipdot(&addr.sin_addr),
		ntohs( addr.sin_port));
/*
* Use listen to establish a queue length.
*/
	if (Listen(lsock,qlen) < 0) {
:1		V1 "Listen(%d,%d) failed [Err %d=%s=%s]",lsock,qlen,Errinfo V;
		Fail;
	}
/*
* Wait for a connection.  When we get one, we fork off a child to
* take care of it, and the main process here goes back to listen for
* another connection.
*/
	then = UnixTime(0);
loop:
	now = UnixTime(0);
	if (timelimit && now - then > timelimit) {
:2		V2 "Time limit of %d seconds expired.",timelimit V;
		Done;
	}
#ifdef SIGIO
	if (use_sigio) {
		int flags;
:4		V4 "SIGIO loop ..." V;
		flags = Fcntl(lsock, F_GETFL, 0);
:5		V5 "flags=%08X=0%0=%d.",flags,flags,flags V;
		i = Fcntl(lsock, F_SETFL, flags | FASYNC);
		i = Fcntl(lsock, F_SETOWN, getpid());
		while (!sigiofl) {
:4			P4 "%s\tSleep for %d sec ...",Vtime(),sigiosleep V;
:9			fprintf(stderr,"sleep(%d) ...\n",sigiosleep); fflush(stderr);
			sleep(sigiosleep);
:9			fprintf(stderr,"sleep(%d) returned.\n",sigiosleep); fflush(stderr);
:4			P4 "%s\tSleep returned.",Vtime() V;
		}
:9		fprintf(stderr,"SIGIO loop done.\n"); fflush(stderr);
	}
#endif
	if (use_accept) {
:4		V4 "Accept ..." V;
		cliens = sizeof(client);
		if ((asock = Accept(lsock,&client,&cliens)) < 0) {
:1			V1 "Accept(%d,%08X,%08X) failed [Err %d=%s=%s]",lsock,&client,&cliens,Errinfo V;
			Fail;
		}
	}
#ifdef SIGIO
	sigiofl = 0;
#endif
:2	V2 "Got connection file=%d from %s.%d.",
:2		asock,
:2		ipdot(&client.sin_addr),
:2		ntohs( client.sin_port) V;
	if (child = Fork()) {
:2		V2 "Started child process %d for %s.%d.",
:2			child,
:2			ipdot(&client.sin_addr),
:2			ntohs( client.sin_port) V;
:5		V5 "Parent process %d closes listening socket %d.",mypid,asock V;
		close(asock);
		asock = -1;
		if (conlim && ++ cons > conlim) {
:2			V2 "Connection limit of %d reached, exiting.",conlim V;
			Done;
		}
		Loop;
	}
:2	Sleep(Vlvl);	/* Give parent a chance to comment and sleep */
	progname = pname = "agent";
	mypid = getpid();
:2	V2 "Child process %d started with accept socket %d.",mypid,asock V;
:5	V5 "Child process %d closes listening socket %d.",mypid,lsock V;
	close(lsock);
	lsock = -1;
:2	V3 "Child process %d for %s.%d.",
:2		mypid,
:2		ipdot(&client.sin_addr),
:2		ntohs( client.sin_port) V;
	if (exec) {
:2		V2 "Dup connection onto stdio files." V;
		Dup2(asock,0);
		Dup2(asock,1);
	/*	Dup2(asock,2);	** Don't do stderr */
	}
	MinStrMF(&Buf,bufmax+1,"Buf");
	MinStrMF(&Cmd,cmdmax+1,"Cmd");
	MinStrMF(&Rsp,rspmax+1,"Rsp");
	Buf.p = Buf.v;
:6	V6 "Buf=[%lX,%d,%d,%lX]",FldsBV(Buf) V;
:6	V6 "Cmd=[%lX,%d,%d,%lX]",FldsBV(Cmd) V;
:6	V6 "Rsp=[%lX,%d,%d,%lX]",FldsBV(Rsp) V;
/*
* Read from the socket
*/
	while (GetCmd()) {
		if (++ msgs  && msglim && msgs > msglim) {
:2			V2 "Message limit of %d reached, exiting.",msglim V;
			Done;
		}
		Msg(&Cmd);
	}
#if 0
	while (1) {
:3		V3 "Select asock=%d.",asock V;
		FD_ZERO(&rmask);
		FD_SET(asock,&rmask);
		if (select(asock+1, &rmask,0,0,0) > 0) {
:3			V3 "Got input, rmask=%08X",rmask V;
			if (!FD_ISSET(asock,&rmask)) {
				V1 "### asock=%d bit not set in rmask ###",asock V;
			}
			if (cmdmax-Cmd.l < 1) {
				cmdmax += 10;
:2				V2 "Increase Cmd buffer to %d bytes.",cmdmax V;
				MinStrMF(&Cmd,cmdmax+1,"Cmd");
			}
:5			V5 "Cmd: [%08lX,%d,%d,%X]",Cmd.v,Cmd.l,Cmd.m,Cmd.p V;
			if ((n = read(asock, Cmd.v+Cmd.l, cmdmax-Cmd.l)) > 0) {
				Cmd.l += n;				/* Total bytes read so far */
				Cmd.v[Cmd.l] = 0;		/* Nul-terminate for debugging */
:3				V3 "Got from %s.%d %s\n",
:3					ipdot(&client.sin_addr),
:3					ntohs( client.sin_port),
:3					DspSV(Cmd) V;
:5				H6(Cmd.v,Cmd.l,"Cmd");
				if (++ msgs  && msglim && msgs > msglim) {
:2					V2 "Message limit of %d reached, exiting.",msglim V;
					Done;
				}
				if (!Index(Cmd.v,'\n')) {
					V3 "Message is incomplete." V;
					continue;
				}
				Msg(&Cmd);
				Cmd.l = 0;
			} else
			if (n == 0) {	/* zero-byte read is EOF */
:2				V2 "EOF on socket %d from %s.%d.",
:2					asock,
:2					ipdot(&client.sin_addr),
:2					ntohs( client.sin_port) V;
				Done;
			} else {
:2				V1 "Error %d=%s=%s on read from %s.%d.",
:2					Errinfo,
:2					ipdot(&client.sin_addr),
:2					ntohs( client.sin_port) V;
				Fail;
			}
		}
	}
#endif
fail:
:2	V2 "Can't read from socket %d [Err %d=%s=%s]",asock,Errinfo V;
done:
	if (lsock >= 0) close(lsock);
	if (asock >= 0) close(asock);
	Exit(r);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Read one command from the client.  We keep reading until a newline.
*/
int GetCmd()
{	int   r=0;
	int   c=0;
	int   e=errno;
	char* ecode="OK";
	int   n;
:f	Fenter("GetCmd");
:6	V6 "Buf=[%lX,%d,%d,%lX]",FldsBV(Buf) V;
:6	V6 "Cmd=[%lX,%d,%d,%lX]",FldsBV(Cmd) V;
	Cmd.l = 0;
	Cmd.p = Cmd.v;
:6	V6 "Cmd=[%lX,%d,%d,%lX]",FldsBV(Cmd) V;
	while (c != '\n') {
		if (Buf.p >= Buf.v + Buf.l) {
			V4 "Read more data into buffer ..." V;
			Buf.p = Buf.v;
			Buf.l = 0;
:6			V6 "before read(%d,%08lX,%d)",asock,Buf.v,Buf.m V;
			n = read(asock,Buf.v,Buf.m);
:6			V6 "after  read(%d,%08lX,%d)=%d [Err %d=%s=%s]",asock,Buf.v,Buf.m,n,Errinfo V;
			if (n > 0) {
				Buf.l += n;
				Buf.v[Buf.l] = 0;
			} else if (n == 0) {
				V2 "EOF on asock=%d.",asock V;
				ecode = "EOF";
				Fail;
			} else {
				V2 "Err %d=%s=%s on asock=%d.",Errinfo,asock V;
				e = errno;
				ecode = sys_errname(errno);
				Fail;
			}
		}
:6		V6 "c=%02X='%c' Buf=%s",B8(c),Dsp(c),Dspp(Buf.p) V;
		c = *Buf.p++;
:6		V6 "c=%02X='%c' Buf=%s",B8(c),Dsp(c),Dspp(Buf.p) V;
		if (Cmd.p >= Cmd.v + Cmd.m) {
:6			V6 "Cmd=[%lX,%d,%d,%lX] expand ...",FldsBV(Cmd) V;
			MinStrMF(&Cmd,Cmd.m+100,"Cmd");
:6			V6 "Cmd=[%lX,%d,%d,%lX] expanded.",FldsBV(Cmd) V;
		}
		*Cmd.p++ = c;
:6		V6 "Cmd=%s",DspSV(Cmd) V;
		r++;
	}
	*Cmd.p = 0;
	Cmd.p = Cmd.v;
:3	V3 "Cmd=[%lX,%d,%d,%lX]",FldsBV(Cmd) V;
:3	V3 "Cmd=%s",DspSV(Cmd) V;
done:
:f	Fexit;
	return r;
fail:
	errno = e;
:f	Fexit;
	return r;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
DmpSKin(p)
	SKin *p;
{
:1	V1 "\tfmly = %d.",p->sin_family V;
:1	V1 "\taddr = %08X=%s.",p->sin_addr,ipdot(&p->sin_addr) V;
:1	V1 "\tport = %04X=%d.",p->sin_port,ntohs(p->sin_port) V;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
int F_int(f)
{	int  r=0;
	CP   p;
	int  c;
:f	Fenter("F_int");
:1	V1 "Decode field %d as int.",f V;
	if (f < msgf) {
		for (p=msgp[f]->p; (c = *p) && isdigit(c); p++) {
			r = (10 * r) + (c - '0');
:5			V5 "c=%c r=%d.",c,r V;
		}
:5		V5 "c=%c r=%d at end of string..",c,r V;
	} else {
		r = 0;
:5		V5 "r=%d default.",c,r V;
	}
:1	V1 "Exit %d.",r V;
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Validate a connection from a client.
*/
int C_conn()
{	int r=0;
	int f;
:f	Fenter("C_conn");
:1	V1 "Command: conn." V;
	for (f=1; f<msgf; f++) {
:3		V3 "Field: %s",DspSP(msgp[f]) V;
	}
	if (msgf > 1 && Streq(msgp[1]->p,myname)) {
:2		V2 "Conn %s accepted.",Dspp(myname) V;
		connected = 1;
	} else if (msgf <= 1) {
:1		V1 "Conn without args." V;
		connected = 0;
	} else {
:1		V1 "Conn %s differs from myname %s",DspSP(msgp[1]),Dspp(myname) V;
		connected = 0;
	}
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
int oct(p,n)
	CP  p;		/* First byte of number */
	int n;		/* Number of digits */
{	int r=0;
	int c;
	while (n-- > 0 && (c = *p++)) {
		if ('0' <= c && c <= '7') {
			r = (8 * r) + (c - '0');
:5			V5 "oct: c='%c' r=0%o.",c,r V;
:3		} else {
:3			V3 "oct: c='%c' ignored.",c V;
		}
	}
	return r;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Open a file for a client.  At present, we just open it and return the file
* handle.  We then forget about it, and trust that the client will remember.
* We might want to keep a record of the remotely-opened files.
*/
int C_open()
{	int r=0;
	int f, x;
	int flgs=0;
	CP  name=0;
	int mode=0666;
:f	Fenter("C_open");
	for (f=1; f<msgf; f++) {
:3		V3 "Field %d: %s",f,DspSP(msgp[f]) V;
		if (msgp[f]->l > 1 && msgp[f]->p[1] == ':') {
			x = *msgp[f]->p;
:6			V6 "'%c' %s",Dsp(x),Dspp(msgp[f]->p+2) V;
			if (x == 'F') {
				flgs = FileFlags(msgp[f]);
:3				V3 "flgs=0x%X.",flgs V;
			} else if (x == 'M') {
				mode = oct(msgp[f]->p+2,msgp[f]->l-2);
:3				V3 "mode=0%o.",mode V;
			} else if (x == 'P') {
				name = msgp[f]->p;
:3				V3 "path=%s",Dspp(name) V;
:2			} else {
:2				V2 "Field %d: %s ignored.",f,DspSP(msgp[f]) V;
			}
		} else {
			name = msgp[f]->p;
:3			V3 "path=%s",Dspp(name) V;
		}
	}
:3	V3 "mode=0%o flgs=0x%X %s",mode,flgs,Dspp(name) V;
	if ((r = open(name,flgs,mode)) >= 0) {
:3		V3 "Opened %d.",r V;
		sprintf(Rsp.v,"<OPEN %d OK\n\0",r);
	} else {
:3		V3 "Can't open %s [Err %d=%s=%s]",Dspp(name),Errinfo V;
		sprintf(Rsp.v,"<OPEN %d %s\n\0",r,sys_errname(errno));
	}
	Rsp.l = strlen(Rsp.v);
:3	V3 "Rsp: %s",DspSV(Rsp) V;
	write(asock,Rsp.v,Rsp.l);
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Stat a file for a client.  We extract the stat info and convert it all to
* a machine-independent symbolic form.
*/
int C_stat()
{	int  r=0;
	int  f, x;
	char fld[64];
	STat st;
	CP   name=0;
:f	Fenter("C_stat");
	for (f=1; f<msgf; f++) {
:3		V3 "Field: %s",DspSP(msgp[f]) V;
		name = msgp[f]->p;
	}
:3	V3 "%s",Dspp(name) V;
	if ((r = stat(name,&st)) < 0) {
:3		V3 "Can't stat %s [Err %d=%s=%s]",Dspp(name),Errinfo V;
		sprintf(Rsp.v,"<STAT %d %s\n\0",r,sys_errname(errno));
	} else {
:3	V3 "Got stat info." V;
		Rsp.l = 0;
		AppStrF(&Rsp,"<STAT",5);
		sprintf(fld," mode=%o\0",st.st_mode);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," ino=%u\0",st.st_ino);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," dev=%u\0",st.st_dev);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," rdev=%u\0",st.st_rdev);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," nlink=%u\0",st.st_nlink);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," uid=%u\0",st.st_uid);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," gid=%u\0",st.st_gid);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," size=%u\0",st.st_size);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," atime=%u\0",st.st_atime);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," mtime=%u\0",st.st_mtime);
		AppStrF(&Rsp,fld,-1);
		sprintf(fld," ctime=%u\n\0",st.st_ctime);
		AppStrF(&Rsp,fld,-1);
	}
:3	V3 "Rsp: %s",DspSV(Rsp) V;
	r = write(asock,Rsp.v,Rsp.l);
fail:
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Here we handle a read request from a client.  The message has the fields:
*   >READ
*   file handle
*   byte count
* We read the data into our Dat buffer, and send a message with the info:
*   >READ
*   file handle
*   byte count
* This is followed by a newline, then the data, plus another newline to ensure
* flushing.
*/
int C_read()
{	int r=0;
	int e=0, f, i, n;
:f	Fenter("C_read");
:1	V1 "Command: read." V;
	f = F_int(1);
	n = F_int(2);
:2	V2 "Read %d bytes from file %d.",n,f);
	MinStrMF(&Dat,n+1,"Dat");
	Dat.l = 0;
	while (Dat.l < n) {
		i = read(f,Dat.v+Dat.l,n-Dat.l);
		if (i > 0) {
			Dat.l += i;
:5			V5 "Got %d of %d bytes.",Dat.l,n V;
		}
		if (i == 0) {
:2			V2 "Read from file %d got EOF",f V;
			break;
		} else {
			e = errno;
:2			V2 "Read from file %d failed [Err %d=%s=%s]",f,Errinfo V;
			break;
		}
	}
:5	V5 "Read loop done with %d of %d bytes, e=%d.",Dat.l,n,e V;
	sprintf(Rsp.v,"<READ %d %d\n\0",f,Dat.l);
	Dat.v[Dat.l++] = '\n';	/* Append a newline to ensure flushing */
	Rsp.l = strlen(Rsp.v);	/* Total length includes the newline */
:3	V3 "Rsp: %s",DspSV(Rsp) V;
	write(asock,Rsp.v,Rsp.l);
:3	V3 "Dat: %s",DspSV(Dat) V;
	write(asock,Dat.v,Dat.l);
fail:
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Here we handle a write request from a client.  The message has the fields:
*   >WRITE
*   file handle
*   byte count
*   DATA\n
* We write the data to the file, and send a message with the info:
*   >WRITE
*   file handle
*   byte count
*   ECODE
* In case the write partially succeeds, we will return the actual
* byte count and the error code.  For successful writes, the error
* code will be "OK".
*/
int C_write()
{	int   r=0;
	int   e=0;
	char* ecode = "OK";
	int   f, i, n;
:f	Fenter("C_write");
:1	V1 "Command: write." V;
	f = F_int(1);
	n = F_int(2);
:2	V2 "Get %d bytes from client %d.",n,f);
	MinStrMF(&Dat,n+1,"Dat");
	Dat.l = 0;
	while (Dat.l < n) {
		i = read(asock,Dat.v+Dat.l,n-Dat.l);
		if (i > 0) {
			Dat.l += i;
:5			V5 "Got %d of %d bytes.",Dat.l,n V;
		}
		if (i == 0) {
			e = 0; ecode = "EOF";
:2			V2 "Read from socket %d got EOF",asock V;
			break;
		} else {
			ecode = sys_errname(e = errno);
:2			V2 "Read from socket %d failed [Err %d=%s=%s]",asock,Errinfo V;
			break;
		}
	}
	if ((r = Dat.l) > 0) {
		r = write(f,Dat.v,Dat.l);
		V4 "Wrote %d of %d bytes to file %d.",r,Dat.l,f V;
		if (r < Dat.l) {
			ecode = sys_errname(e = errno);
		}
	} else {
		V2 "No data to write to file %d.",f V;
	}
	sprintf(Rsp.v,"<WRITE %d %d %s\n\0",f,r,ecode);
	Rsp.l = strlen(Rsp.v);	/* Total length includes the newline */
:3	V3 "Rsp: %s",DspSV(Rsp) V;
	write(asock,Rsp.v,Rsp.l);
:3	V3 "Dat: %s",DspSV(Dat) V;
	write(asock,Dat.v,Dat.l);
fail:
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
int C_close()
{	int r=0;
	int f;
:f	Fenter("C_close");
:1	V1 "Command: close." V;
	f = F_int(1);
:2	V2 "Close %d file %d.",f);
	errno = 0;
	r = close(f);
	sprintf(Rsp.v,"<CLOSE %d %d %s\n\0",f,r,sys_errname(errno));
	Rsp.l = strlen(Rsp.v);	/* Total length includes the newline */
:3	V3 "Rsp: %s",DspSV(Rsp) V;
	write(asock,Rsp.v,Rsp.l);
fail:
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
int C_exit()
{	int r=0;
	int f;
:f	Fenter("C_exit");
:1	V1 "Command: exit." V;
	f = F_int(1);
:2	V2 "exit %d file %d.",f);
	errno = 0;
	sprintf(Rsp.v,"<CLOSE %d %d %s\n\0",f,r,sys_errname(errno));
	Rsp.l = strlen(Rsp.v);	/* Total length includes the newline */
:3	V3 "Rsp: %s",DspSV(Rsp) V;
	write(asock,Rsp.v,Rsp.l);
	exit(f);
fail:
	V1 "### Can't exit(%d) ###",f V;
:f	Fexit;
	return r;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/

#define CmdTbl struct cmd_tbl
CmdTbl {
	char* sym;
	int (*fp)();
};
CmdTbl cmdtbl[] = {
	{">CONN",	C_conn},
	{">STAT",	C_stat},
	{">OPEN",	C_open},
	{">READ",	C_read},
	{">WRITE",	C_write},
	{">CLOSE",	C_close},
	{">EXIT",	C_exit},
	{0,0}
};

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Here we process one incoming message.  This entails splitting it into fields,
* and calling a function that corresponds to the first field. The function will
* handle sending any replies.
*/
Msg(bp)
	Str* bp;
{	int  r=0;
	int  f, i;
	int  n;
	byte c;
	CP   p;
	CP   q;
:f	Fenter("Msg");
:3	V3 "%s",DspSP(bp) V;
	for (msgf=i=0,p=q=bp->v; i<n; i++) {
		c = bp->v[i];
:6		V6 "c=%02X='%c'",B8(c),Dsp(c) V;
		if ((c == ' ') || (c == '\r') || (c == '\n')) {
			q = bp->v+i;
			*q = 0;
			if (q-p > 0) {
:4				V4 "Field %d: \"%s\"",msgf,p V;
				if (msgf >= msgm) {
:5					V5 "Message has %d fields, must be increased.",msgm V;
:5					H5(msgp,msgm*sizeof(Str*),"msgp");
					msgp = (Str**)SizChunk(msgp,msgm*sizeof(Str*),(msgm+2)*sizeof(Str*),"msgp");
:5					V5 "msgp=%08lX size=%d.",msgp,(msgm+2)*sizeof(Str*) V;
					if (!msgp) {
:1						V1 "Can't get %d bytes for message list.",(msgm+2)*sizeof(Str*) V;
						Exit(ENOMEM);
					}
					msgm += 2;
:3					V3 "Message list now holds %d messages.",msgm V;
:5					H5(msgp,msgm*sizeof(Str*),"msgp");
				}
				if (!(msgp[msgf] = MakStr(msgp[msgf],p,q-p))) {
:1					V1 "Can't build msgp[%d] string (%d=%s=%s)",msgf,Errinfo);
					Exit(ENOMEM);
				}
:3				V3 "Field %d: %s",msgf,DSP(msgp[msgf]) V;
				++msgf;
:5			} else {
:5				V5 "Null field ignored." V;
			}
			p = q + 1;
		}
	}
:2	for (i=0; i<msgf; i++) {
:2		V2 "Message field %2d: %s",i,DSP(msgp[i]) V;
:2	}
	cmd = msgp[0]->p;
	Rsp.l = 0;
:2	V2 "Message command: %s",Dspp(cmd) V;
	for (i=0; cmdtbl[i].sym; i++ ) {
		if (!strcasecmp(cmd,cmdtbl[i].sym)) {
			V2 "Recognized command %s",Dspp(cmd) V;
			r = cmdtbl[i].fp();
		}
	}
:f	Fexit;
	return r;
}
