#include "D_D.h"
#include "mem.h"
#include "str.h"
#include "srvr.h"
#include "sys_socket.h"
#include "sys_in.h"
#include "sys_netdb.h"

global Srvr* srvrs = 0;	/* List of known servers */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* CALL:
*   Srvr* daSrvrLink(name)
*   	Str*  name;	-- Name of server
*
* DESCRIPTION:
*
* This returns a connection to a server.  It first checks to see if we
* have  a  connection, and if so, returns it.  If not, we go through a
* number of attempts to make a connection (aka association).
*
* If successful, we return a  pointer  to  the  Srvr  struct  for  the
* server; failure returns a null pointer.
*
* AUTHOR: John Chambers.
*
* CHANGES:
*/
FCT	Srvr* daSrvrLink(name)
	Str*  name;		/* Name of server */
{	Srvr* sp=0;		/* Returned server */
	Srvr* xp=0;		/* Unused server slot */
	int   port=0;
	HOST *hp, *gethostbyname();
:f	Fenter("daSrvrLink");
:5	V5D "Called for %s",DspSP(name) V;

	for (sp = srvrs; sp; sp = sp->next) {
		if (!CmpStrM(&sp->name,name,"name")) {
:5			V5D "Already talking to server %s",DspSP(name) V;
			return sp;
		} else if (!xp && !sp->name.l) {
			xp = sp;	/* Note unused srvr slot */
		}
	}
:5	V5D "Not yet talking to server %s",DspSP(name) V;
	if (xp) {
		sp = xp;
	} else {
		GetChunkF(sp,Srvr*,sizeof(Srvr),"Srvr");
		sp->next = srvrs;
		srvrs = sp;
	}
	DupStrM(&sp->name,name,"Srvr");

	if (!port) port = 4217;
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Create socket on which to read.                                      *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
:6	V6 "before socket(AF_INET=%d, SOCK_STREAM=%d, 0)",AF_INET,SOCK_STREAM V;
	sp->sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sp->sock < 0) {
:1		V1 "### Can't get STREAM socket [Err %d=%s=%s]",Errinfo V;
		sp = 0;
		Fail;
	}
:3	V3 "Got TCP stream socket %d.",sp->sock V;
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Construct address of socket to send to.   Gethostbyname()  retuns  a *
* structure  including the network address of the specified host.  The *
* port number is taken from the command line.                          *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
:5	V5D "Get host info for %s ...",DspSV(sp->name) V;
	hp = Gethostbyn(sp->name.v);
	Bcopy(hp->h_addr,&(sp->addr.sin_addr.s_addr),hp->h_length);
	sp->addr.sin_family = AF_INET;
	sp->addr.sin_port = htons(port);
:6	H6(&sp->addr,sizeof(sp->addr),"addr");

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Try to make a connection to the server.                              *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
	if (Connect(sp->sock,&sp->addr,sizeof(sp->addr)) < 0) {
:2		V1 "Can't connect to %s.%u.",
:2			ipdot(&sp->addr.sin_addr),
:2			ntohs( sp->addr.sin_port) V;
		Fail;
	}
:3	V3 "Connected to %s.%u.",
:3		ipdot(&sp->addr.sin_addr),
:3		ntohs( sp->addr.sin_port) V;
	Done;

fail:
:2	V2D "Can't contact %s",DspSP(name) V;
done:
:f	Fexit;
	return sp;
}
